/******************************************************************************
 * Spine Runtimes License Agreement
 * Last updated July 28, 2023. Replaces all prior versions.
 *
 * Copyright (c) 2013-2024, Esoteric Software LLC
 *
 * Integration of the Spine Runtimes into software or otherwise creating
 * derivative works of the Spine Runtimes is permitted under the terms and
 * conditions of Section 2 of the Spine Editor License Agreement:
 * http://esotericsoftware.com/spine-editor-license
 *
 * Otherwise, it is permitted to integrate the Spine Runtimes into software or
 * otherwise create derivative works of the Spine Runtimes (collectively,
 * "Products"), provided that each user of the Products must obtain their own
 * Spine Editor license and redistribution of the Products in any form must
 * include this license and copyright notice.
 *
 * THE SPINE RUNTIMES ARE PROVIDED BY ESOTERIC SOFTWARE LLC "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL ESOTERIC SOFTWARE LLC BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES,
 * BUSINESS INTERRUPTION, OR LOSS OF USE, DATA, OR PROFITS) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THE
 * SPINE RUNTIMES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

using System;
using Unity.Collections;
using UnityEngine;
using UnityEngine.Profiling;
using UnityEngine.UIElements;
using UIVertex = UnityEngine.UIElements.Vertex;

namespace Spine.Unity {
	public class SpineVisualElement : VisualElement {
		public new class UxmlFactory : UxmlFactory<SpineVisualElement, UxmlTraits> { }

		public new class UxmlTraits : VisualElement.UxmlTraits {
			UxmlStringAttributeDescription m_StartingAnimation = new UxmlStringAttributeDescription { name = "starting-animation" };

			UxmlStringAttributeDescription m_InitialSkinName = new UxmlStringAttributeDescription { name = "initial-skin-name" };
			UxmlBoolAttributeDescription m_StartingLoop = new UxmlBoolAttributeDescription { name = "starting-loop" };
			UxmlFloatAttributeDescription m_TimeScale = new UxmlFloatAttributeDescription { name = "time-scale", defaultValue = s_DefaultTimeScale };

			public override void Init (VisualElement ve, IUxmlAttributes bag, CreationContext cc) {
				base.Init(ve, bag, cc);

				SpineVisualElement element = (SpineVisualElement)ve;

				string animation = null;
				if (m_StartingAnimation.TryGetValueFromBag(bag, cc, ref animation))
					element.startingAnimation = animation;


				string skinName = null;
				if (m_InitialSkinName.TryGetValueFromBag(bag, cc, ref skinName))
					element.initialSkinName = skinName;


				bool loop = false;
				if (m_StartingLoop.TryGetValueFromBag(bag, cc, ref loop))
					element.startingLoop = loop;

				element.timeScale = m_TimeScale.GetValueFromBag(bag, cc);
			}
		}

		const float s_DefaultTimeScale = 1.0f;

		static CustomStyleProperty<SkeletonDataAsset> s_AssetStyleProperty = new("--skeleton-asset");

		SkeletonRendererInstruction currentInstructions = new();// to match existing code better
		protected Spine.Unity.MeshGeneratorUIElements meshGenerator = new MeshGeneratorUIElements();

		SkeletonDataAsset m_SkeletonDataAsset;
		AnimationState state;
		Skeleton skeleton;
		IVisualElementScheduledItem m_ScheduledItem;
		float scale = 100;
		Bounds referenceMeshBounds;
		bool adjustToBounds = true;
		float offsetX, offsetY;
		VisualElement m_RendererElement;

		/// <summary>Update mode to optionally limit updates to e.g. only apply animations but not update the mesh.</summary>
		public UpdateMode UpdateMode { get { return updateMode; } set { updateMode = value; } }
		protected UpdateMode updateMode = UpdateMode.FullUpdate;

		public bool freeze { get; set; }

		public bool unscaledTime { get; set; }

		public float timeScale { get; set; } = s_DefaultTimeScale;

		public string initialSkinName { get; set; }

		public string startingAnimation { get; set; } = "";

		public bool startingLoop { get; set; } = true;

		bool IsValid { get { return skeleton != null; } }

		public SkeletonDataAsset skeletonDataAsset {
			get => m_SkeletonDataAsset;
			set {
				if (m_SkeletonDataAsset == value)
					return;

				m_SkeletonDataAsset = value;
				Initialize(false);
			}
		}



		public AnimationState animationState {
			get {
				Initialize(false);
				return state;
			}
		}

		public SpineVisualElement () {
			RegisterCallback<CustomStyleResolvedEvent>(OnCustomStylesResolved);
			RegisterCallback<DetachFromPanelEvent>(OnDetatchedCallback);


			m_RendererElement = new VisualElement();
			m_RendererElement.generateVisualContent += GenerateVisualContents;
			m_RendererElement.pickingMode = PickingMode.Ignore;
			m_RendererElement.style.position = Position.Absolute;
			m_RendererElement.style.top = 0;
			m_RendererElement.style.left = 0;
			m_RendererElement.style.bottom = 0;
			m_RendererElement.style.right = 0;
			Add(m_RendererElement);

			// TODO: if this ever becomes a property that can change, implement propery deregistration
			if (adjustToBounds) {
				m_RendererElement.RegisterCallback<GeometryChangedEvent>(OnGeometryChanged);
			}
		}

		void OnGeometryChanged (GeometryChangedEvent evt) {
			AdjustOffsetScaleToMeshBounds(m_RendererElement);
		}

		void OnDetatchedCallback (DetachFromPanelEvent evt) {
			ClearElement();
		}

		public void ClearElement () {
			skeleton = null;
			DisposeUISubmeshes();
		}

		public virtual void Update () {

#if UNITY_EDITOR
			if (!Application.isPlaying) {
				Update(0f);
				return;
			}
#endif

			if (freeze) return;
			Update(unscaledTime ? Time.unscaledDeltaTime : Time.deltaTime);

			m_RendererElement.MarkDirtyRepaint();
		}

		public virtual void Update (float deltaTime) {
			if (!this.IsValid) return;

			if (updateMode < UpdateMode.OnlyAnimationStatus)
				return;
			UpdateAnimationStatus(deltaTime);

			if (updateMode == UpdateMode.OnlyAnimationStatus)
				return;
			ApplyAnimation();
		}

		protected void UpdateAnimationStatus (float deltaTime) {
			deltaTime *= timeScale;
			//skeleton.Update(deltaTime);
			state.Update(deltaTime);
		}

		protected void ApplyAnimation () {

			if (updateMode != UpdateMode.OnlyEventTimelines)
				state.Apply(skeleton);
			else
				state.ApplyEventTimelinesOnly(skeleton);

			skeleton.UpdateWorldTransform();
		}


		void Initialize (bool overwrite) {
			if (this.IsValid && !overwrite) return;

			if (this.skeletonDataAsset == null) return;
			var skeletonData = this.skeletonDataAsset.GetSkeletonData(false);
			if (skeletonData == null) return;

			if (skeletonDataAsset.atlasAssets.Length <= 0 || skeletonDataAsset.atlasAssets[0].MaterialCount <= 0) return;

			this.state = new Spine.AnimationState(skeletonDataAsset.GetAnimationStateData());
			if (state == null) {
				Clear();
				return;
			}

			this.skeleton = new Skeleton(skeletonData) {
				ScaleX = 1,
				ScaleY = -1
			};

			// Set the initial Skin and Animation
			if (!string.IsNullOrEmpty(initialSkinName))
				skeleton.SetSkin(initialSkinName);

			if (!string.IsNullOrEmpty(startingAnimation)) {
				var animationObject = skeletonDataAsset.GetSkeletonData(false).FindAnimation(startingAnimation);
				if (animationObject != null) {
					state.SetAnimation(0, animationObject, startingLoop);
				}
			}

			if (adjustToBounds) {
				AdjustReferenceMeshBounds();
			}

			if (m_ScheduledItem == null)
				m_ScheduledItem = schedule.Execute(Update).Every(1);

			m_RendererElement.MarkDirtyRepaint();
		}

		protected class UISubmesh {
			public NativeArray<UIVertex>? vertices = null;
			public NativeArray<ushort>? indices = null;

			public NativeSlice<UIVertex> verticesSlice;
			public NativeSlice<ushort> indicesSlice;
		}

		readonly ExposedList<UISubmesh> uiSubmeshes = new ExposedList<UISubmesh>();

		protected void PrepareUISubmeshCount (int targetCount) {
			int oldCount = uiSubmeshes.Count;
			uiSubmeshes.EnsureCapacity(targetCount);
			for (int i = oldCount; i < targetCount; ++i) {
				uiSubmeshes.Add(new UISubmesh());
			}
		}

		protected void PrepareUISubmesh (UISubmesh uiSubmesh, int vertexCount, int indexCount) {
			bool shallReallocateVertices = uiSubmesh.vertices == null || uiSubmesh.vertices.Value.Length < vertexCount;
			if (shallReallocateVertices) {
				int allocationCount = vertexCount;
				if (uiSubmesh.vertices != null) {
					allocationCount = Math.Max(vertexCount, 2 * uiSubmesh.vertices.Value.Length);
					uiSubmesh.vertices.Value.Dispose();
				}
				uiSubmesh.vertices = new NativeArray<UIVertex>(allocationCount, Allocator.Persistent, NativeArrayOptions.ClearMemory);
			}
			if (shallReallocateVertices || uiSubmesh.verticesSlice.Length != vertexCount) {
				uiSubmesh.verticesSlice = new NativeSlice<UIVertex>(uiSubmesh.vertices.Value, 0, vertexCount);
			}

			bool shallReallocateIndices = uiSubmesh.indices == null || uiSubmesh.indices.Value.Length < indexCount;
			if (shallReallocateIndices) {
				int allocationCount = indexCount;
				if (uiSubmesh.indices != null) {
					allocationCount = Math.Max(indexCount, uiSubmesh.indices.Value.Length * 2);
					uiSubmesh.indices.Value.Dispose();
				}
				uiSubmesh.indices = new NativeArray<ushort>(allocationCount, Allocator.Persistent, NativeArrayOptions.ClearMemory);
			}
			if (shallReallocateIndices || uiSubmesh.indicesSlice.Length != indexCount) {
				uiSubmesh.indicesSlice = new NativeSlice<ushort>(uiSubmesh.indices.Value, 0, indexCount);
			}
		}

		protected void DisposeUISubmeshes () {
			for (int i = 0, count = uiSubmeshes.Count; i < count; ++i) {
				UISubmesh uiSubmesh = uiSubmeshes.Items[i];
				if (uiSubmesh.vertices != null) uiSubmesh.vertices.Value.Dispose();
				if (uiSubmesh.indices != null) uiSubmesh.indices.Value.Dispose();
			}
			uiSubmeshes.Clear();
		}

		protected void GenerateVisualContents (MeshGenerationContext context) {
			if (!this.IsValid) return;


			MeshGeneratorUIElements.GenerateSkeletonRendererInstruction(currentInstructions, skeleton, null,
				null,
				 false,
				false);

			int submeshCount = currentInstructions.submeshInstructions.Count;
			PrepareUISubmeshCount(submeshCount);
			// Generate meshes.
			for (int i = 0; i < submeshCount; i++) {
				var submeshInstructionItem = currentInstructions.submeshInstructions.Items[i];
				UISubmesh uiSubmesh = uiSubmeshes.Items[i];

				meshGenerator.Begin();
				meshGenerator.AddSubmesh(submeshInstructionItem);
				PrepareUISubmesh(uiSubmesh, meshGenerator.VertexCount, meshGenerator.SubmeshIndexCount(0));

				meshGenerator.FillVertexData(ref uiSubmesh.verticesSlice);
				meshGenerator.FillTrianglesSingleSubmesh(ref uiSubmesh.indicesSlice);

				var submeshMaterial = submeshInstructionItem.material;

				Texture usedTexture = submeshMaterial.mainTexture;

				FillContext(context, uiSubmesh, usedTexture);
			}
		}

		public void AdjustReferenceMeshBounds () {

			// Need one update to obtain valid mesh bounds
			Update(0.0f);
			MeshGeneratorUIElements.GenerateSkeletonRendererInstruction(currentInstructions, skeleton,
				null, null, false, false);
			int submeshCount = currentInstructions.submeshInstructions.Count;
			meshGenerator.Begin();

			for (int i = 0; i < submeshCount; i++) {
				var submeshInstructionItem = currentInstructions.submeshInstructions.Items[i];
				meshGenerator.AddSubmesh(submeshInstructionItem);
			}

			referenceMeshBounds = meshGenerator.GetMeshBounds();
		}

		void AdjustOffsetScaleToMeshBounds (VisualElement visualElement) {
			Rect targetRect = visualElement.layout;
			float xScale = targetRect.width / referenceMeshBounds.size.x;
			float yScale = targetRect.height / referenceMeshBounds.size.y;
			this.scale = Math.Min(xScale, yScale);
			float targetOffsetX = targetRect.width / 2;
			float targetOffsetY = targetRect.height / 2;
			this.offsetX = targetOffsetX - referenceMeshBounds.center.x * this.scale;
			this.offsetY = targetOffsetY - referenceMeshBounds.center.y * this.scale;

			visualElement.style.translate = new StyleTranslate(new Translate(offsetX, offsetY, 0));
			visualElement.style.transformOrigin = new TransformOrigin(0, 0, 0);
			visualElement.style.scale = new Scale(new Vector3(scale, scale, 1));
		}

		void FillContext (MeshGenerationContext context, UISubmesh submesh, Texture texture) {
			MeshWriteData meshWriteData = context.Allocate(submesh.verticesSlice.Length, submesh.indicesSlice.Length, texture);

			meshWriteData.SetAllVertices(submesh.verticesSlice);
			meshWriteData.SetAllIndices(submesh.indicesSlice);
		}

		void OnCustomStylesResolved (CustomStyleResolvedEvent evt) {
			SkeletonDataAsset specifiedAsset;
			if (evt.customStyle.TryGetValue(s_AssetStyleProperty, out specifiedAsset)) {
				skeletonDataAsset = specifiedAsset;
				m_RendererElement.MarkDirtyRepaint();
			}
		}
	}
}
